class Node:
    def __init__(self, id_, addr):
        self._id = id_
        self._addr = addr

    @property
    def id(self):
        return self._id

    @property
    def addr(self):
        return self._addr



import asyncio
from collections import namedtuple
from datetime import datetime

from aiokrpc import KRPCServer
from aiokrpc.exceptions import KRPCErrorResponse
from aiokrpc.exceptions import KRPCProtocolError

from .routing_table import RoutingTable
from .schemas import ANNOUNCE_PEER_ARGS
from .schemas import ANNOUNCE_PEER_ARGS_REMOTE
from .schemas import ANNOUNCE_PEER_RESULT
from .schemas import ANNOUNCE_PEER_RESULT_REMOTE
from .schemas import FIND_NODE_ARGS
from .schemas import FIND_NODE_ARGS_REMOTE
from .schemas import FIND_NODE_RESULT
from .schemas import FIND_NODE_RESULT_REMOTE
from .schemas import GET_PEERS_ARGS
from .schemas import GET_PEERS_ARGS_REMOTE
from .schemas import GET_PEERS_RESULT
from .schemas import GET_PEERS_RESULT_REMOTE
from .schemas import PING_ARGS
from .schemas import PING_ARGS_REMOTE
from .schemas import PING_RESULT
from .schemas import PING_RESULT_REMOTE
from .utils import calc_sha1
from .utils import call_timeout
from .utils import decode_info_hash
from .utils import random
from .utils import run_every

PeerInfo = namedtuple("peer_info", ["addr", "port", "implied_port", "added"])


class ArgsError(Exception):
    pass


class ResultError(Exception):
    pass


class DHT(KRPCServer):
    def __init__(self, local_id, **kwargs):
        super().__init__(**kwargs)

        self.id = local_id
        self.routing_table = RoutingTable(local_id)

        self.torrents = {}
        self.salts = []

        # region Callbacks registration
        self.register_callback(self.ping, arg_schema=PING_ARGS, result_schema=PING_RESULT)
        self.register_callback(self.find_node, arg_schema=FIND_NODE_ARGS, result_schema=FIND_NODE_RESULT)
        self.register_callback(self.get_peers, arg_schema=GET_PEERS_ARGS, result_schema=GET_PEERS_RESULT)
        self.register_callback(self.announce_peer, arg_schema=ANNOUNCE_PEER_ARGS, result_schema=ANNOUNCE_PEER_RESULT)
        # endregion

    # region Server methods
    def ping(self, addr, id):
        self._run_future(self._add_or_update_node(id, addr))

        return self._get_result_id()

    def find_node(self, addr, id, target):
        self._run_future(self._add_or_update_node(id, addr))

        return {
            "nodes": self.routing_table[target],
            **self._get_result_id()
        }

    def get_peers(self, addr, id, info_hash):
        self._run_future(self._add_or_update_node(id, addr))

        peers = self.torrents.get(info_hash, None)
        if peers:
            return {
                **self._get_result_id(), **self._get_result_token(addr),
                "values": [(peer.addr[0], peer.addr[1] if peer.implied_port else peer.port) for peer in peers]
            }
        else:
            return {**self.find_node(addr, id, info_hash), **self._get_result_token(addr)}

    def announce_peer(self, addr, id, info_hash, port, token, implied_port=0):
        self._run_future(self._add_or_update_node(id, addr))

        if self._check_token(addr, token):
            self.torrents.setdefault(info_hash, set()).add(
                PeerInfo(
                    addr=addr,
                    port=port,
                    implied_port=implied_port,
                    added=datetime.now()
                )
            )
            return self._get_result_id()
        else:
            raise KRPCProtocolError("Bad token")

    # endregion

    # region Utils
    def _get_result_id(self):
        return {"id": self.id}

    def _gen_tokens(self, addr):
        for salt in self.salts:
            yield calc_sha1(b"".join((bytes(str(addr), "utf-8"), salt)))

    def _gen_token(self, addr):
        return next(self._gen_tokens(addr))

    def _get_result_token(self, addr):
        return {"token": self._gen_token(addr)}

    def _check_token(self, addr, token):
        return any(token == t for t in self._gen_tokens(addr))

    def _add_node(self, node_id, addr):
        self.routing_table.add(node_id, addr)

    def _run_every(self, f, delay):
        self._run_future(run_every(f, delay))

    @staticmethod
    def server_version():
        return "AD"

    # endregion

    # region Remote calls
    async def _remote_call(self, addr, method, kwargs, timeout=3, arg_schema=None, result_schema=None):
        def _args_error(e):
            raise ArgsError()

        def _result_error(e):
            raise ResultError()

        try:
            result = await call_timeout(
                self.call_remote(
                    addr, method, **self.apply_schema(kwargs, arg_schema or {}, _args_error)
                ), timeout, None
            )
            if result:
                return result[0], self.apply_schema(result[1], result_schema or {}, _result_error)

            return None
        except (KRPCErrorResponse, ResultError):
            return None

    async def remote_ping(self, addr, timeout=3):
        return await self._remote_call(
            addr, "ping",
            kwargs={"id": self.id},
            timeout=timeout,
            arg_schema=PING_ARGS_REMOTE,
            result_schema=PING_RESULT_REMOTE
        )

    async def remote_find_node(self, addr, target_id, timeout=3):
        return await self._remote_call(
            addr, "find_node",
            kwargs={"id": self.id, "target": target_id},
            timeout=timeout,
            arg_schema=FIND_NODE_ARGS_REMOTE,
            result_schema=FIND_NODE_RESULT_REMOTE
        )

    async def remote_get_values(self, addr, info_hash, timeout=3):
        return await self._remote_call(
            addr, "get_peers",
            kwargs={"id": self.id, "info_hash": info_hash},
            timeout=timeout,
            arg_schema=GET_PEERS_ARGS_REMOTE,
            result_schema=GET_PEERS_RESULT_REMOTE
        )

    async def remote_announce_peer(self, addr, info_hash, port, token, implied_port, timeout=3):
        return await self._remote_call(
            addr, "announce_peer",
            kwargs={"id": self.id, "info_hash": info_hash, "port": port, "token": token, "implied_port": implied_port},
            timeout=timeout,
            arg_schema=ANNOUNCE_PEER_ARGS_REMOTE,
            result_schema=ANNOUNCE_PEER_RESULT_REMOTE
        )

    async def _add_or_update_node(self, id_, addr):
        if (id_, addr) in self.routing_table:
            self._add_node(id_, addr)
        else:
            response = await self.remote_ping(addr)
            if response:
                self._add_node(response[1]["id"], addr)

    # endregion

    def connection_made(self):
        for args in (
                (self._refresh_nodes, 60),
                (self._rotate_salts, 60),
                (self._forget_torrents, 60)):
            self._run_every(*args)

    # region Periodic tasks
    async def _refresh_nodes(self):
        responses = filter(
            lambda response: response,
            await asyncio.gather(
                *(self.remote_ping(node[1]) for node in self.routing_table.enum_nodes_for_refresh()),
                loop=self.loop
            )
        )

        for addr, data in responses:
            self._add_node(node_id=data["id"], addr=addr)

    def _rotate_salts(self):
        self.salts.insert(0, calc_sha1(random(128)))

        if len(self.salts) > 10:
            self.salts.pop(-1)

    def _forget_torrents(self):
        now = datetime.now()

        for info_hash, values in list(self.torrents.items()):
            for value in list(values):
                if (now - value.added).seconds >= 30 * 60:
                    values.remove(value)

            if not values:
                self.torrents.pop(info_hash)

    # endregion

    async def _group_invoke(self, cb, peers):
        return filter(
            lambda response: response and response[1]["id"] != self.id,
            await asyncio.gather(
                *(cb(peer) for peer in peers),
                loop=self.loop
            )
        )

    async def _get_values(self, info_hash, announce=False, port=None):
        result = set()
        known = set()
        tokens = set()
        peers = map(lambda it: it[1], self.routing_table[info_hash])

        while True:
            responses = await self._group_invoke(lambda peer: self.remote_get_values(peer, info_hash), peers)

            candidates = set()
            for addr, data in responses:
                candidates.update(data.get("nodes", []))
                result.update(data.get("values", []))

                token = data.get("token")
                if token:
                    tokens.add((addr, token))

            closest = RoutingTable.get_k_closest(info_hash, candidates - known, key=lambda it: it[0])

            if closest:
                known.update(candidates)
                peers = map(lambda it: it[1], closest)
            else:
                if announce:
                    await self._group_invoke(
                        lambda it: self.remote_announce_peer(
                            it[0], info_hash, port or 0, it[1], 1 if port is None else 0),
                        tokens
                    )
                    return
                else:
                    return result

    async def bootstrap(self, initial_peers):
        known = set()
        peers = initial_peers

        while True:
            responses = await self._group_invoke(lambda peer: self.remote_find_node(peer, self.id), peers)

            candidates = set()
            for addr, data in responses:
                self._add_node(node_id=data["id"], addr=addr)
                candidates.update(data["nodes"])

            closest = RoutingTable.get_k_closest(self.id, candidates - known, key=lambda it: it[0])

            if closest:
                known.update(candidates)
                peers = map(lambda it: it[1], closest)
            else:
                break

    async def announce(self, info_hash, port=None):
        await self._get_values(decode_info_hash(info_hash), announce=True, port=port)

    def __getitem__(self, item):
        return self._get_values(decode_info_hash(item))



from .bucket import Bucket
from .node import Node


class RoutingTable:
    def __init__(self, local_id):
        self._local_id = local_id
        self._buckets = {Bucket(0, 2 ** 160)}

    def add(self, id_, addr):
        for bucket in self._buckets:
            if bucket.id_in_range(id_):
                added = bucket.add(Node(id_, addr))

                if not added and self._split(bucket):
                    return self.add(id_, addr)

    # region Internal methods
    def _split(self, bucket):
        if bucket.range_max - bucket.range_min < bucket.max_capacity:
            return False

        self._buckets.remove(bucket)

        median = (bucket.range_max + bucket.range_min) >> 1  # Divide by half

        for new_bucket in (Bucket(bucket.range_min, median), Bucket(median, bucket.range_max)):
            self._buckets.add(new_bucket)
            for node in bucket.nodes:
                if new_bucket.id_in_range(node.id):
                    new_bucket.add(node)

        return True

    def _get_closest(self, target_id, k=8):
        return self.get_k_closest(
            target_id,
            ((node.id, node.addr) for bucket in self._buckets for node in bucket.nodes),
            key=lambda it: it[0],
            k=k
        )

    # endregion

    def enum_nodes_for_refresh(self):
        for b in self._buckets:
            yield from b.enum_nodes_for_refresh()

    @staticmethod
    def get_k_closest(target, iterable, key=None, k=8):
        if key and not callable(key):
            raise

        return sorted(iterable, key=lambda n: (key(n) if key else n) ^ target)[:k]

    def __getitem__(self, item):
        # Item -- node_id or tuple (node_id, k), by default `k == 8`
        if isinstance(item, int):
            return self._get_closest(item)
        elif isinstance(item, tuple) and len(item) == 2:
            return self._get_closest(*item)
        else:
            raise TypeError("Unsupported type")

    def __contains__(self, item):
        return any(item in bucket for bucket in self._buckets)



from .utils import decode_id
from .utils import decode_nodes
from .utils import decode_peers
from .utils import encode_id
from .utils import encode_nodes
from .utils import encode_peers

_ID_ENCODE_SCHEMA = {"type": "binary", "minlength": 20, "maxlength": 20, "coerce": encode_id}
_ID_DECODE_SCHEMA = {"type": "integer", "coerce": decode_id}

_INFO_HASH_ENCODE_SCHEMA = _ID_ENCODE_SCHEMA
_INFO_HASH_DECODE_SCHEMA = _ID_DECODE_SCHEMA

_NODES_ENCODE_SCHEMA = {"type": "binary", "coerce": encode_nodes}
_NODES_DECODE_SCHEMA = {"type": "list", "coerce": decode_nodes}

_VALUES_ENCODE_SCHEMA = {"type": "list", "coerce": encode_peers, "schema": {"type": "binary"}}
_VALUES_DECODE_SCHEMA = {"type": "list", "coerce": decode_peers}

_TOKEN_SCHEMA = {"type": "binary"}

PING_ARGS = {"id": {"required": True, **_ID_DECODE_SCHEMA}}
PING_RESULT = {"id": {"required": True, **_ID_ENCODE_SCHEMA}}
PING_ARGS_REMOTE = {"id": {"required": True, **_ID_ENCODE_SCHEMA}}
PING_RESULT_REMOTE = {"id": {"required": True, **_ID_DECODE_SCHEMA}}

FIND_NODE_ARGS = {
    "id": {"required": True, **_ID_DECODE_SCHEMA},
    "target": {"required": True, **_ID_DECODE_SCHEMA}}
FIND_NODE_RESULT = {
    "id": {"required": True, **_ID_ENCODE_SCHEMA},
    "nodes": {"required": True, **_NODES_ENCODE_SCHEMA}}
FIND_NODE_ARGS_REMOTE = {
    "id": {"required": True, **_ID_ENCODE_SCHEMA},
    "target": {"required": True, **_ID_ENCODE_SCHEMA}}
FIND_NODE_RESULT_REMOTE = {
    "id": {"required": True, **_ID_DECODE_SCHEMA},
    "nodes": {"required": True, **_NODES_DECODE_SCHEMA}}

GET_PEERS_ARGS = {
    "id": {"required": True, **_ID_DECODE_SCHEMA},
    "info_hash": {"required": True, **_INFO_HASH_DECODE_SCHEMA}}
GET_PEERS_RESULT = {
    "id": {"required": True, **_ID_ENCODE_SCHEMA},
    "token": {"type": "binary", "required": True},
    "nodes": _NODES_ENCODE_SCHEMA,
    "values": _VALUES_ENCODE_SCHEMA
}
GET_PEERS_ARGS_REMOTE = {
    "id": {"required": True, **_ID_ENCODE_SCHEMA},
    "info_hash": {"required": True, **_INFO_HASH_ENCODE_SCHEMA}}
GET_PEERS_RESULT_REMOTE = {
    "id": {"required": True, **_ID_DECODE_SCHEMA},
    "token": {"type": "binary", "required": True},
    "nodes": _NODES_DECODE_SCHEMA,
    "values": _VALUES_DECODE_SCHEMA
}

ANNOUNCE_PEER_ARGS = {
    "id": {"required": True, **_ID_DECODE_SCHEMA},
    "info_hash": {"required": True, **_INFO_HASH_DECODE_SCHEMA},
    "token": {"required": True, **_TOKEN_SCHEMA},
    "implied_port": {"type": "integer", "required": False, "min": 0, "max": 1}}
ANNOUNCE_PEER_RESULT = {"id": {"required": True, **_ID_ENCODE_SCHEMA}}
ANNOUNCE_PEER_ARGS_REMOTE = {
    "id": {"required": True, **_ID_ENCODE_SCHEMA},
    "info_hash": {"required": True, **_INFO_HASH_ENCODE_SCHEMA},
    "token": {"required": True, **_TOKEN_SCHEMA},
    "implied_port": {"type": "integer", "required": False, "min": 0, "max": 1}}
ANNOUNCE_PEER_RESULT_REMOTE = {"id": {"required": True, **_ID_DECODE_SCHEMA}}



from datetime import datetime


class NodeStat:
    def __init__(self):
        self._added = datetime.now()
        self._last_response = self._added

    def renew(self):
        self._last_response = datetime.now()

    @property
    def rate(self):
        now = datetime.now()
        delta = (now - self._last_response).seconds

        if delta < 15 * 60:
            return (now - self._last_response).seconds / (30 * 60) + 1
        elif delta > 30 * 60:
            return -1
        else:
            return 0

    @property
    def last_response(self):
        return self._last_response



from setuptools import setup, find_packages

with open("README.md") as f:
    long_description = f.read()

setup(
    name="aio-btdht",
    version="0.0.8",
    description="Asyncio Bittorrent DHT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=["Programming Language :: Python"],
    keywords="Async Bittorrent-DHT DHT",
    author="D.Bashkirtsevich",
    author_email="bashkirtsevich@gmail.com",
    url="https://github.com/bashkirtsevich-llc/aiobtdht",
    license="GPL3 License",
    include_package_data=True,
    zip_safe=True,
    packages=find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.6.*",
    install_requires=[
        "aio-krpc-server==0.0.6"
    ]
)


import asyncio
from hashlib import sha1
from os import getrandom
from socket import inet_aton, inet_ntoa


def random(size=1):
    return getrandom(size)


def calc_sha1(data):
    return sha1(data).digest()


def encode_id(id_):
    return id_.to_bytes(20, "big")


def decode_id(id_):
    if len(id_) != 20:
        raise ValueError("Wrong length")

    return int.from_bytes(id_, "big")


def decode_info_hash(info_hash):
    return decode_id(info_hash)


def encode_addr(addr):
    host, port = addr
    return b"".join((inet_aton(host), port.to_bytes(2, "big")))


def decode_addr(addr):
    host, port = addr[:4], addr[4:6]
    return (inet_ntoa(host), int.from_bytes(port, "big"))


def encode_nodes(nodes):
    return b"".join(b"".join((encode_id(id_), encode_addr(addr))) for id_, addr in nodes)


def decode_nodes(nodes):
    if len(nodes) % 26 != 0:
        raise ValueError("Wrong length")

    return [(decode_id(nodes[i: i + 20]), decode_addr(nodes[i + 20: i + 26])) for i in range(0, len(nodes), 26)]


def encode_peers(peers):
    return [encode_addr(peer) for peer in peers]


def decode_peers(peers):
    return [decode_addr(peer) for peer in peers]


async def run_every(f, delay):
    while True:
        r = f()

        if asyncio.iscoroutine(r):
            await r

        await asyncio.sleep(delay)


async def call_timeout(f, timeout, default):
    try:
        return await asyncio.wait_for(f, timeout)
    except asyncio.TimeoutError:
        return default



from .node_stat import NodeStat


class Bucket:
    def __init__(self, range_min, range_max, max_capacity=8):
        self._range_min = range_min
        self._range_max = range_max
        self._max_capacity = max_capacity
        self._nodes = {}

    def __hash__(self):
        return hash((self._range_min, self._range_max))

    def id_in_range(self, id_):
        return self._range_min <= id_ < self._range_max

    def add(self, node):
        if not self.id_in_range(node.id):
            raise IndexError("Node id not in bucket range")

        if node in self._nodes:
            self._nodes[node].renew()
            return True
        elif len(self._nodes) < self._max_capacity:
            self._nodes[node] = NodeStat()
            return True
        else:
            can_delete = list(filter(lambda it: it[1].rate < 0, self._enum_nodes()))
            if can_delete:
                for it in can_delete:
                    self._nodes.pop(it)

                return self.add(node)
            else:
                return False

    def _enum_nodes(self):
        for node, stat in self._nodes.items():
            yield node, stat

    def enum_nodes_for_refresh(self):
        return map(lambda it: it[0].addr, filter(lambda it: it[1].rate <= 0, self._enum_nodes()))

    @property
    def range_min(self):
        return self._range_min

    @property
    def range_max(self):
        return self._range_max

    @property
    def max_capacity(self):
        return self._max_capacity

    @property
    def nodes(self):
        # Return only good nodes
        return map(lambda it: it[0], filter(lambda it: it[1].rate > 0, self._enum_nodes()))

    def __contains__(self, item):
        if isinstance(item, int):
            return any(item == node.id for node in self._nodes)
        elif isinstance(item, tuple):
            return any(item == (node.id, node.addr) for node in self._nodes)
        else:
            raise TypeError("Unsupported type")
